import cv2
import numpy as np
import pyclipper
from shapely.geometry import Polygon


class SegDetectorRepresenter:
    def __init__(self, thresh=0.3, box_thresh=0.5, max_candidates=1000, unclip_ratio=2.0):
        self.min_size = 3
        self.thresh = thresh
        self.box_thresh = box_thresh
        self.max_candidates = max_candidates
        self.unclip_ratio = unclip_ratio
    
    def __call__(self, pred, height, width):
        """
        batch: (image, polygons, ignore_tags
        batch: a dict produced by dataloaders.
            image: tensor of shape (N, C, H, W).
            polygons: tensor of shape (N, K, 4, 2), the polygons of objective regions.
            ignore_tags: tensor of shape (N, K), indicates whether a region is ignorable or not.
            shape: the original shape of images.
            filename: the original filenames of images.
        pred:
            binary: text region segmentation map, with shape (N, H, W)
            thresh: [if exists] thresh hold prediction with shape (N, H, W)
            thresh_binary: [if exists] binarized with threshhold, (N, H, W)
        """

        pred = pred[0, :, :]
        segmentation = self.binarize(pred)
        
        boxes, scores = self.boxes_from_bitmap(pred, segmentation, width, height)
        
        return boxes, scores
    
    def binarize(self, pred):
        return pred > self.thresh
    
    def boxes_from_bitmap(self, pred, bitmap, dest_width, dest_height):
        """
        _bitmap: single map with shape (H, W),
            whose values are binarized as {0, 1}
        """
        
        assert len(bitmap.shape) == 2
        # bitmap = _bitmap.cpu().numpy()  # The first channel
        # pred = pred.cpu().detach().numpy()
        height, width = bitmap.shape
        # print(bitmap)
        # cv2.imwrite("./test/output/test.jpg",(bitmap * 255).astype(np.uint8))
        contours, _ = cv2.findContours((bitmap * 255).astype(np.uint8), cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
        num_contours = min(len(contours), self.max_candidates)
        boxes = np.zeros((num_contours, 4, 2), dtype=np.int16)
        scores = np.zeros((num_contours,), dtype=np.float32)
        rects = []
        for index in range(num_contours):
            contour = contours[index].squeeze(1)
            points, sside = self.get_mini_boxes(contour)
            if sside < self.min_size:
                continue
            points = np.array(points)
            score = self.box_score_fast(pred, contour)
            if self.box_thresh > score:
                continue
            # print(points)
            box = self.unclip(points, unclip_ratio=self.unclip_ratio).reshape(-1, 1, 2)
            # print(box)
            box, sside = self.get_mini_boxes(box)
            if sside < self.min_size + 2:
                continue
            box = np.array(box)
            if not isinstance(dest_width, int):
                dest_width = dest_width.item()
                dest_height = dest_height.item()
            
            box[:, 0] = np.clip(np.round(box[:, 0] / width * dest_width), 0, dest_width)
            box[:, 1] = np.clip(np.round(box[:, 1] / height * dest_height), 0, dest_height)
            boxes[index, :, :] = box.astype(np.int16)
            scores[index] = score
        return boxes, scores
    
    def unclip(self, box, unclip_ratio=1.5):
        poly = Polygon(box)
        
        distance = poly.area * unclip_ratio / (poly.length  )
        offset = pyclipper.PyclipperOffset()
        offset.AddPath(box, pyclipper.JT_ROUND, pyclipper.ET_CLOSEDPOLYGON)
        expanded = np.array(offset.Execute(distance))
        return expanded
    
    def get_mini_boxes(self, contour):
        bounding_box = cv2.minAreaRect(contour)
        points = sorted(list(cv2.boxPoints(bounding_box)), key=lambda x: x[0])
        
        index_1, index_2, index_3, index_4 = 0, 1, 2, 3
        if points[1][1] > points[0][1]:
            index_1 = 0
            index_4 = 1
        else:
            index_1 = 1
            index_4 = 0
        if points[3][1] > points[2][1]:
            index_2 = 2
            index_3 = 3
        else:
            index_2 = 3
            index_3 = 2
        
        box = [points[index_1], points[index_2], points[index_3], points[index_4]]
        return box, min(bounding_box[1])
    
    def box_score_fast(self, bitmap, _box):
        h, w = bitmap.shape[:2]
        box = _box.copy()
        xmin = np.clip(np.floor(box[:, 0].min()).astype(np.int), 0, w - 1)
        xmax = np.clip(np.ceil(box[:, 0].max()).astype(np.int), 0, w - 1)
        ymin = np.clip(np.floor(box[:, 1].min()).astype(np.int), 0, h - 1)
        ymax = np.clip(np.ceil(box[:, 1].max()).astype(np.int), 0, h - 1)
        
        mask = np.zeros((ymax - ymin + 1, xmax - xmin + 1), dtype=np.uint8)
        box[:, 0] = box[:, 0] - xmin
        box[:, 1] = box[:, 1] - ymin
        cv2.fillPoly(mask, box.reshape(1, -1, 2).astype(np.int32), 1)
        return cv2.mean(bitmap[ymin:ymax + 1, xmin:xmax + 1], mask)[0]



import onnxruntime as ort
import numpy as np
import time
import traceback
from multiprocessing.dummy import Pool as ThreadPool
import cv2
import os

cur_path = os.path.dirname(os.path.abspath(__file__))


class PPrecPredictor:
    '''
    百度的识别模型
    '''
    def __init__(self, model_path = os.path.join(cur_path,"models/pprec_v2.onnx") ,
                 target_h = 32 ,
                 batch_num = 1,
                 keys_txt_path=os.path.join(cur_path,"models/ppocr_keys_v1.txt"),
                 in_names="x",
                 out_names=["save_infer_model/scale_0.tmp_1"]):
        self.keys_txt_path = keys_txt_path
        self.in_names = in_names
        self.out_names = out_names        
        self.target_h = target_h
        self.batch_num = batch_num
        self.sess = ort.InferenceSession(model_path)
 
        self.character_str = []
        character_dict_path = keys_txt_path
        use_space_char = True
        if character_dict_path is None:
            self.character_str = "0123456789abcdefghijklmnopqrstuvwxyz"
            dict_character = list(self.character_str)
        else:
            with open(character_dict_path, "rb") as fin:
                lines = fin.readlines()
                for line in lines:
                    line = line.decode('utf-8').strip("\n").strip("\r\n")
                    self.character_str.append(line)
            if use_space_char:
                self.character_str.append(" ")
            dict_character = list(self.character_str)

        dict_character = self.add_special_char(dict_character)
        self.dict = {}
        for i, char in enumerate(dict_character):
            self.dict[char] = i
        self.character = dict_character

    def preprocess(self,im,max_wh_ratio = 3):
        target_h = self.target_h
        scale = im.shape[0] * 1.0 / target_h
        w = im.shape[1] / scale
        w = int(w)
        img = cv2.resize(im,(w, target_h),interpolation=cv2.INTER_AREA) # ,interpolation=cv2.INTER_AREA 在这里效果最好

        if self.batch_num > 1:
            # 最大
            max_resized_w = int(max_wh_ratio * target_h)
            padding_im = np.zeros((target_h,max_resized_w,3), dtype=np.float32)
            padding_im[:,0:w,:] = img
            img = padding_im

        img -= 127.5
        img /= 127.5

        image = img.transpose(2, 0, 1)
        transformed_image = np.expand_dims(image, axis=0)
        return transformed_image

    def npbox2box(self,npbox):
        '''
        numpy格式转列表格式
        '''
        x = npbox[0][0]
        y = npbox[0][1]
        w = npbox[2][0] - x + 1
        h = npbox[2][1] - y + 1
        return x,y,w,h

    def check_edge(self,x,y,bb_w,bb_h,h,w):
        '''
        边界检查
        '''
        x = max(0,x)
        y = max(0,y)
        bb_w = w-bb_w-x if x + bb_w > w else bb_w
        bb_h = h-bb_h-y if y + bb_h > h else bb_h
        return x,y,bb_w,bb_h

    def expand(self,x,y,bb_w,bb_h,h,w,delta=1):
        '''
        外扩delta个像素
        '''
        x -= delta
        y -= delta 
        bb_w += 2*delta
        bb_h += 2*delta
        x,y,bb_w,bb_h = self.check_edge(x,y,bb_w,bb_h,h,w)
        return x,y,bb_w,bb_h

    def add_special_char(self, dict_character):
        dict_character = ['blank'] + dict_character
        return dict_character

    def get_rotate_crop_image(self,img, points):
        img_height, img_width = img.shape[0:2]
        left = int(np.min(points[:, 0]))
        right = int(np.max(points[:, 0]))
        top = int(np.min(points[:, 1]))
        bottom = int(np.max(points[:, 1]))
        img_crop = img[top:bottom, left:right, :].copy()
        dst_img = img_crop
        return dst_img


    def decode(self, text_index, text_prob=None, is_remove_duplicate=False):
        """ convert text-index into text-label. """
        result_list = []
        ignored_tokens = self.get_ignored_tokens()
        batch_size = len(text_index)
        th = 0.15
        for batch_idx in range(batch_size):
            selection = np.ones(len(text_index[batch_idx]), dtype=bool)
            if is_remove_duplicate:
                selection[1:] = text_index[batch_idx][1:] != text_index[
                    batch_idx][:-1]
            for ignored_token in ignored_tokens:
                selection &= text_index[batch_idx] != ignored_token
            # 不使用置信度阈值过滤
            # char_list = [
            #     self.character[text_id]
            #     for idx,text_id in enumerate(text_index[batch_idx][selection])
            # ]
            # 使用置信度阈值过滤
            char_list = [
                self.character[text_id]
                for idx,text_id in enumerate(text_index[batch_idx][selection])
                if text_prob[batch_idx][selection][idx] > th
            ]

            if text_prob is not None:
                conf_list = text_prob[batch_idx][selection]
            else:
                conf_list = [1] * len(selection)
            if len(conf_list) == 0:
                conf_list = [0]

            text = ''.join(char_list)
            result_list.append((text, np.mean(conf_list).tolist()))
        return result_list

    def get_ignored_tokens(self):
        return [0]  # for ctc blank

    def predict_rbg(self, im):
        """
        预测
        """
        t1 = time.time()
        preds = self.sess.run(self.out_names, {self.in_names: im.astype(np.float32)}) # 12ms
        # print("run cost: ",time.time()-t1)
        preds = preds[0]
        # preds = softmax(preds)
        preds_idx = preds.argmax(axis=2)
        preds_prob = preds.max(axis=2)
        
        text = self.decode(preds_idx, preds_prob, is_remove_duplicate=True)

        return text
    
    def pp_predict(self,data):
        '''
        
        '''
        im,box = data
        x,y,bb_w,bb_h = box 
        # 外扩delta个像素
        # x,y,bb_w,bb_h = self.expand(x,y,bb_w,bb_h,im.shape[0],im.shape[1],2)
        # box = np.array([[x,y],[x+bb_w,y],[x+bb_w,y+bb_h],[x,y+bb_h]])
        try:
            # 裁剪
            partImg_array = self.get_rotate_crop_image(im, box.astype(np.float32))
            partImg = self.preprocess(partImg_array.astype(np.float32))
            result = self.predict_rbg(partImg)  ##识别的文本
        except Exception as e:
            print(traceback.format_exc())
            result = [("",0)]
        simPred,prob = result[0]
        return self.npbox2box(box),simPred,prob

    def __call__(self,im,texts_list,use_mp = False, process_num = 1):
        
        """
        crnn模型，ocr识别
        @@model,
        @@converter,
        @@im:Array
        @@text_recs:text box
        @@ifIm:是否输出box对应的img
        """

        results = []
        if not use_mp:
            # 不用多线程
            for box in texts_list:
                box_ = self.npbox2box(box)
                if  not (box_[2] > 3 and box_[3] > 3 and box_[2] / box_[3] < 50):
                    continue
                # x,y,bb_w,bb_h = box 
                # box = np.array([[x,y],[x+bb_w,y],[x+bb_w,y+bb_h],[x,y+bb_h]])
                # 裁剪
                partImg_array = self.get_rotate_crop_image(im, box.astype(np.float32))
                partImg = self.preprocess(partImg_array.astype(np.float32))
                try:
                    result = self.predict_rbg(partImg)  ##识别的文本
                except Exception as e:
                    print(traceback.format_exc())
                    continue
                simPred,prob = result[0]
                results.append([self.npbox2box(box),simPred,prob])
        else:
            # 多线程方法二（更快）
            datas = [(im,box) for box in texts_list if self.npbox2box(box)[2] > 3 and self.npbox2box(box)[3] > 3 and self.npbox2box(box)[2] / self.npbox2box(box)[3] < 50]
            pool = ThreadPool(processes = process_num)
            results = pool.map(self.pp_predict, datas)
            pool.close()
            pool.join()
            used_boxes = set([self.npbox2box(box[1]) for box in datas])         # 做了识别的框
            text_boxes = set([self.npbox2box(box) for box in texts_list])    # 所有框
            rest_boxes = list(text_boxes - used_boxes)          # 剩余未做识别的框
            results.extend([(box,"",0) for box in rest_boxes])

        return results



from pp_ocr import pp_det
from pp_ocr import pp_rec
import cv2
import numpy as np
import time

# 初始化模型
print("init models...")
pp_detor = pp_det.DBNET()
pp_recor = pp_rec.PPrecPredictor()

short_size = 960


def predict(img_path,use_mp = True ,process_num = 10):
    img = cv2.imdecode(np.fromfile(img_path,dtype=np.uint8),-1)
    
    if len(img.shape) == 3 and img.shape[2] > 3:
        img = img[:,:,:3]
     
    start_time = time.time()
    boxes_list, score_list = pp_detor.process(np.asarray(img).astype(np.uint8),short_size=short_size)
    print("det nums:{} cost time: {:.4f}".format(len(boxes_list),time.time()-start_time))

    start_time = time.time()
    result = pp_recor(np.array(img), boxes_list,use_mp= use_mp,process_num= process_num)
    print("rec nums:{} cost time: {:.4f}".format(len(result),time.time()-start_time))

    return result



import onnxruntime as rt
import numpy as np
import time
import traceback
from multiprocessing.dummy import Pool as ThreadPool
import cv2
import os
from pp_ocr.decode import SegDetectorRepresenter

cur_path = os.path.dirname(os.path.abspath(__file__))

class SingletonType(type):
    def __init__(cls, *args, **kwargs):
        super(SingletonType, cls).__init__(*args, **kwargs)

    def __call__(cls, *args, **kwargs):
        obj = cls.__new__(cls, *args, **kwargs)
        cls.__init__(obj, *args, **kwargs)
        return obj


class DBNET(metaclass=SingletonType):
    def __init__(self, MODEL_PATH = os.path.join(cur_path,"models/pp_det.onnx")):
        self.sess = rt.InferenceSession(MODEL_PATH)
        self.decode_handel = SegDetectorRepresenter()
        self.mean = (0.485, 0.456, 0.406)
        self.std = (0.229, 0.224, 0.225)


    def process(self, img, short_size):
        t1 = time.time()
        # img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        h, w = img.shape[:2]
        if h > w:
            scale_h = short_size / h
            tar_w = w * scale_h
            tar_w = max(int(round(tar_w / 32) * 32), 32)
            scale_w = tar_w / w

        else:
            scale_w = short_size / w
            tar_h = h * scale_w
            tar_h = max(int(round(tar_h / 32) * 32), 32)
            scale_h = tar_h / h

        img = cv2.resize(img, None, fx=scale_w, fy=scale_h)
        # img = cv2.resize(img,(320,320))

        img = img.astype(np.float32)

        img /= 255.0
        img -= self.mean
        img /= self.std
        img = img.transpose(2, 0, 1)
        transformed_image = np.expand_dims(img, axis=0)
        print("transforme cost: ",time.time()- t1)

        t1 = time.time()
        # dbnet推理
        # out = self.sess.run(["out"], {"input": transformed_image.astype(np.float32)})
        # PPOCR_det推理
        out = self.sess.run(["sigmoid_0.tmp_0"], {"x": transformed_image.astype(np.float32)})
        print("run cost: ",time.time()- t1)

        t1 = time.time()
        box_list, score_list = self.decode_handel(out[0][0], h, w)
        print("decode cost: ",time.time()- t1)
        if len(box_list) > 0:
            idx = box_list.reshape(box_list.shape[0], -1).sum(axis=1) > 0  # 去掉全为0的框
            box_list, score_list = box_list[idx], score_list[idx]
        else:
            box_list, score_list = [], []
        return box_list, score_list


import pathlib
from setuptools import setup, find_packages


APPNAME = "pp_ocr"
VERSION = "1.2"

with open("requirements.txt") as fin:
    REQUIRED_PACKAGES = fin.read()

setup(
    name = APPNAME,
    version = VERSION,
    description="paddleocr识别",                                      # 长描述，用于在Pypi上面显示
    author="xdg",
    license="Apache Software License",
    packages=find_packages(),                                       # 包模块
    install_requires=REQUIRED_PACKAGES,
    include_package_data=True,
   )
